/*
 * @FilePath: /output_framebuffer/output_framebuffer.c
 * @Version: 1.0.0
 * @Author: zys
 * @LastAuthor   : zys
 * @CreationDate: 2020-02-17 21:11:39
 * @LastEditTime : 2020-03-22 21:48:30
 * @Description :  
 */


/*----------------------------------------------------------------------------*/
/* change log ----------------------------------------------------------------*/



/*----------------------------------------------------------------------------*/
/* Includes ------------------------------------------------------------------*/
#include <arpa/inet.h>
#include <errno.h>
#include <getopt.h>
#include <pthread.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <syslog.h>
#include <unistd.h>

#include <linux/types.h> /* for videodev2.h */
#include <linux/videodev2.h>

#include <fcntl.h>
#include <linux/fb.h>
#include <sys/mman.h>


#include <jpeglib.h>
#include <setjmp.h>


#include "../../mjpg_streamer.h"
#include "../../utils.h"


/* Private define ------------------------------------------------------------*/
#define OUTPUT_PLUGIN_NAME "Framebuffer output plugin"
#define BUFFER_SIZE        1024
#define MAX_FRAME_SIZE     (256 * 1024)
#define TEN_K              (10 * 1024)


#define IMAGEWIDTH  800
#define IMAGEHEIGHT 480

/* Private macro -------------------------------------------------------------*/
/* Private typedef -----------------------------------------------------------*/
typedef struct
{
    unsigned char blue;
    unsigned char green;
    unsigned char red;
} __attribute__((packed)) PIXEL; //颜色模式RGB


/* 错误处理相关 */
struct my_error_mgr {
    struct jpeg_error_mgr pub;

    jmp_buf setjmp_buffer;
};


typedef struct my_error_mgr* my_error_ptr;

/* Private function prototypes -----------------------------------------------*/
void* worker_thread(void*);
void  worker_cleanup(void*);

/* Private variables ---------------------------------------------------------*/

// -mjpg
static pthread_t worker;
static globals*  pglobal;
// static unsigned char *frame = NULL;
static int input_number = 0;

// --fb
static int           xres           = 0;
static int           yres           = 0;
static int           bits_per_pixel = 0;
static char*         fb_addr        = NULL;
static char*         fb_buf1        = NULL;
static int           fb_size;
static unsigned char frame_buffer[IMAGEWIDTH * IMAGEHEIGHT * 3];

/* Global  variables ---------------------------------------------------------*/
/* Private functions ---------------------------------------------------------*/



/******************************************************************************
Description.: print help for this plugin to stdout
Input Value.: -
Return Value: -
******************************************************************************/
void help(void)
{
    fprintf(stderr, " ---------------------------------------------------------------\n"
                    " Help for output plugin..: " OUTPUT_PLUGIN_NAME "\n"
                    " ---------------------------------------------------------------\n"
                    " The following parameters can be passed to this plugin:\n\n"
                    " [-w | --www ]...........: folder that contains webpages in \n"
                    "                           flat hierarchy (no subfolders)\n"
                    " [-p | --port ]..........: TCP port for this HTTP server\n"
                    " [-c | --credentials ]...: ask for \"username:password\" on connect\n"
                    " [-n | --nocommands ]....: disable execution of commands\n"
                    " ---------------------------------------------------------------\n");
}

/*** plugin interface functions ***/
/******************************************************************************
Description.: Initialize this plugin.
              parse configuration parameters,
              store the parsed values in global variables
Input Value.: All parameters to work with.
              Among many other variables the "param->id" is quite important -
              it is used to distinguish between several server instances
Return Value: 0 if everything is OK, other values signal an error
******************************************************************************/
int output_init(output_parameter* param, int id)
{
    int i;


    OPRINT("\n-----%s INIT-----\n\n", OUTPUT_PLUGIN_NAME);

    pglobal               = param->global;
    pglobal->out[id].name = malloc((1 + strlen(OUTPUT_PLUGIN_NAME)) * sizeof(char));
    sprintf(pglobal->out[id].name, "%s", OUTPUT_PLUGIN_NAME);
    DBG("OUT plugin %d name: %s\n", id, pglobal->out[id].name);

    param->argv[0] = OUTPUT_PLUGIN_NAME;

    /* show all parameters for DBG purposes */
    for (i = 0; i < param->argc; i++) {
        DBG("argv[%d]=%s\n", i, param->argv[i]);
    }

    reset_getopt();
    while (1) {
        int                  c              = 0;
        static struct option long_options[] = {
            { "help", no_argument, NULL, 'h' },
            { "input", required_argument, NULL, 'i' },
            { NULL, 0, NULL, 0 }
        };

        c = getopt_long(param->argc, param->argv, "hi:", long_options, NULL);

        /* no more options to parse */
        if (c == -1)
            break;

        switch (c) {
        case 'i':
            input_number = atoi(optarg);
            return 1;
            break;

        case 'h': /* fall through */
        default:
            help();
            return 1;
            break;
        }
    }

    if (!(input_number < pglobal->incnt)) {
        OPRINT("ERROR: the %d input_plugin number is too much only %d plugins loaded\n", input_number, pglobal->incnt);
    }

    OPRINT("input plugin.....: %d: %s\n", input_number, pglobal->in[input_number].plugin);

    return 0;
}




/******************************************************************************
Description.: this will stop the server thread, client threads
              will not get cleaned properly, because they run detached and
              no pointer is kept. This is not a huge issue, because this
              funtion is intended to clean up the biggest mess on shutdown.
Input Value.: id determines which server instance to send commands to
Return Value: always 0
******************************************************************************/
int output_stop(int id)
{
    DBG("will cancel %s thread #%02d\n", OUTPUT_PLUGIN_NAME, id);
    pthread_cancel(worker);

    return 0;
}

/******************************************************************************
Description.: This creates and starts the server thread
Input Value.: id determines which server instance to send commands to
Return Value: always 0
******************************************************************************/
int output_run(int id)
{
    DBG("\nlaunching %s thread #%02d\n\n", OUTPUT_PLUGIN_NAME, id);

    /* create thread and pass context to thread function */
    pthread_create(&worker, NULL, worker_thread, NULL);
    pthread_detach(worker);

    return 0;
}

/******************************************************************************
Description.: This is just an example function, to show how the output
              plugin could implement some special command.
              If you want to control some GPIO Pin this is a good place to
              implement it. Dont forget to add command types and a mapping.
Input Value.: cmd is the command type
              id determines which server instance to send commands to
Return Value: 0 indicates success, other values indicate an error
******************************************************************************/
int output_cmd(int plugin, unsigned int control_id, unsigned int group, int value)
{
    DBG("command (%d, value: %d) for group %d triggered for plugin instance #%02d\n", control_id, value, group, plugin);
    return 0;
}



int fbfd;
struct fb_var_screeninfo vinfo;   /*记录用户可修改的显示控制参数*/
struct fb_fix_screeninfo finfo;    /*取得缓冲区的物理地址*/

int fb_init(void)
{
    // int                      fbfd, fb_size;
    // struct fb_var_screeninfo vinfo; /*记录用户可修改的显示控制参数*/
    // struct fb_fix_screeninfo finfo; /*取得缓冲区的物理地址*/

    int                      fb_size;

    // 打开/dev/fb*设备文件
    fbfd = open("/dev/fb0", O_RDWR | O_CREAT, S_IRWXU);
    if (fbfd < 0) {
        printf("open error\n");
        exit(1);
    }

    //需要的 屏幕内存空间的字节数	,取得显示屏幕参数
    if (ioctl(fbfd, FBIOGET_VSCREENINFO, &vinfo)) {
        printf("Error reading variable information.\n");
        exit(1);
    }

    vinfo.yres_virtual = vinfo.yres * 2;
    vinfo.yoffset = 0;

    if (ioctl(fbfd, FBIOPUT_VSCREENINFO, &vinfo)) {
        printf("Error put variable information.\n");
        exit(1);
    }



    if (ioctl(fbfd, FBIOGET_FSCREENINFO, &finfo)) {
        printf("Error reading variable information.\n");
        exit(1);
    }
  printf("R:%d,G:%d,B:%d \n", vinfo.red.length, vinfo.green.length, vinfo.blue.length );  


    printf("%dx%d, %dbpp\n", vinfo.xres, vinfo.yres, vinfo.bits_per_pixel);

    xres           = vinfo.xres; /*可见解析度*/
    yres           = vinfo.yres;
    bits_per_pixel = vinfo.bits_per_pixel;                     /* 每像素位数*/
    // fb_size        = xres * yres * ((bits_per_pixel + 7) / 8); /*计算屏幕的总大小（字节)*/
    fb_size        = vinfo.yres * finfo.line_length;
    printf("yres_virtual\t: %d \n", vinfo.yres_virtual);
    printf("smem_len\t: %d \n", finfo.smem_len);
    printf("fbsize\t\t: %d \n", fb_size);

    // fd的内存地址空间映射到我们的程序地址空间
    if ((fb_addr = (char*)mmap(0, finfo.smem_len, PROT_READ | PROT_WRITE, MAP_SHARED, fbfd, 0)) < 0) {
        printf("fb mmap error\n");
        exit(1);
    }

    // close(fbfd);

    fb_buf1 = (char*)calloc(1, fb_size);

    if (fb_buf1 == NULL) {
        /* code */

        printf("fb buf1 calloc err!\n");
    }


    printf("fb init ok!\n\n");

    return fb_size;
}




METHODDEF(void)
my_error_exit(j_common_ptr cinfo)
{
    /* cinfo->err really points to a my_error_mgr struct, so coerce pointer */
    my_error_ptr myerr = (my_error_ptr)cinfo->err;

    /* Always display the message. */
    /* We could postpone this until after returning, if we chose. */
    (*cinfo->err->output_message)(cinfo);

    /* Return control to the setjmp point */
    longjmp(myerr->setjmp_buffer, 1);
}


int jpeg_to_rgb24(unsigned char* pdest, unsigned char* psrc, unsigned long insize, unsigned int* pwidth, unsigned int* pheight)
{
    struct jpeg_decompress_struct cinfo;
    struct my_error_mgr           jerr;
    // int image_width, image_height;

    /* 从源码中可以知道这是一个指向指针的指针 */
    JSAMPARRAY buffer = NULL; /* 一行图形数据存储的地方 */
    int        row_stride;    /* 一行数据的字节数 */

    // fprintf(stderr, " DBG(%s, %s(), %d): ", __FILE__, __FUNCTION__, __LINE__);

    /* 1.分配和初始化cinfo */

    /* 绑定错误处理函数 */
    cinfo.err           = jpeg_std_error(&jerr.pub);
    jerr.pub.error_exit = my_error_exit;

    /* 错误处理 */
    if (setjmp(jerr.setjmp_buffer)) {
        printf("not jpeg file!!!\n");
        jpeg_destroy_decompress(&cinfo);
        return -1;
    }

    /* 初始化cinfo */
    jpeg_create_decompress(&cinfo);

    /* 2.初始化cinfo */
    // jpeg_stdio_src(&cinfo, infile);
    // ----从内存读
    jpeg_mem_src(&cinfo, psrc, insize);

    /* 3.读取文件头 */
    jpeg_read_header(&cinfo, TRUE);

    /* 4.开始解压 */
    jpeg_start_decompress(&cinfo);

    /* 记录下图像的宽度和高度，给图像数据存储区分配内存 */
    *pwidth  = cinfo.output_width;
    *pheight = cinfo.output_height;

    DBG("read jpeg size: w%d, h%d\n", cinfo.output_width, cinfo.output_height);

    /* 自己定义的图像存储区 */
    // char *p_image_buf;
    // p_image_buf = (unsigned char *)malloc(abs(image_width * image_height * 3));

    /* 一行字节数 = 行宽 * 每个像素点字节数 */
    row_stride = cinfo.output_width * cinfo.output_components;

    /* 给libjpeg所需要的行缓存分配内存空间 */
    buffer = (*cinfo.mem->alloc_sarray)((j_common_ptr)&cinfo, JPOOL_IMAGE, row_stride, 1);

    /* 5.一行一行地扫描图像 */
    while (cinfo.output_scanline < cinfo.output_height) {
        /* 读取一行图像 */
        jpeg_read_scanlines(&cinfo, buffer, 1);

        /* 将一行图像数据保存起来 */
        // memcpy(p_image_buf + ((cinfo.output_scanline - 1) * row_stride), buffer[0], row_stride); /* 特别注意buffer[0] */

        memcpy(pdest + ((cinfo.output_scanline - 1) * row_stride), buffer[0], row_stride); /* 特别注意buffer[0] */
    }

    /* 6.结束解压 */
    jpeg_finish_decompress(&cinfo);

    /* 7.释放内存 */
    jpeg_destroy_decompress(&cinfo);

    return 0;
}



void fb_show_jpeg(char* pbuf, int fb_size, unsigned int width, unsigned int height)
{
    int                   line_x, line_y;
    long int              location = 0;
    static unsigned short b, g, r;
    static unsigned short rgb; /*2个字节*/
    PIXEL                 pix;
    int                   pixcnt = 0;
    char*                 psrc   = pbuf;

    pixcnt = 0;


    // --每次处理一个像素
    for (line_y = 0; line_y < height; line_y++) {
        for (line_x = 0; line_x < width; line_x++) {

            //-- ok
            // bmp 是按BGR格式。
            // pix.blue  = psrc[pixcnt++];
            // pix.green = psrc[pixcnt++];
            // pix.red   = psrc[pixcnt++];

            // jpg 是按RGB格式。
            pix.red   = psrc[pixcnt++];
            pix.green = psrc[pixcnt++];
            pix.blue  = psrc[pixcnt++];


            r   = ((pix.red >> 3) << 11) & 0xf800;
            g   = ((pix.green >> 2) << 5) & 0x07e0;
            b   = (pix.blue >> 3) & 0x001f;
            rgb = r | g | b;

            // bmp
            // location = line_x*2+(IMAGEHEIGHT - 1 - line_y)*xres*2;

            // jpeg
            location = line_x * 2 + (line_y)*xres * 2;

            //--先放到缓存里面
            *((unsigned short*)(fb_buf1 + location)) = rgb;
        }
    }

    // --更新整屏
    // memcpy(fb_addr, fb_buf1, fb_size);
}


// ok 双缓冲测试成功，没有撕裂了！
void framebuffer_buf_update(int fb_size)
{
	int ret;
	static int buf_index = 0;  // 0-buf0-online


	if (buf_index)
	{
		// --更新整屏 offline

		vinfo.yoffset = vinfo.yres;
		memcpy(fb_addr + vinfo.yoffset * finfo.line_length, fb_buf1, vinfo.yres * finfo.line_length);

		ret = ioctl(fbfd, FBIOPAN_DISPLAY, &vinfo);	

		if (ret)
		{
			printf("----FBIOPAN_DISPLAY-failed: %d \n", ret);
		}

		// printf("fb online buf : %d \n", buf_index);

		buf_index = 0;
	}
	else
	{
		// --更新整屏 offline
		memcpy(fb_addr, fb_buf1, vinfo.yres * finfo.line_length);

		vinfo.yoffset = 0;
		ret = ioctl(fbfd, FBIOPAN_DISPLAY, &vinfo);	

		if (ret)
		{
			printf("----FBIOPAN_DISPLAY-failed: %d \n", ret);
		}

		// printf("fb online buf : %d \n", buf_index);

		buf_index = 1;
	}	
}



void load_jpeg_to_framebuffer(unsigned char* jpegdatabuf, int size)
{
    unsigned int w, h;

    jpeg_to_rgb24(frame_buffer, jpegdatabuf, size, &w, &h);
    fb_show_jpeg((char*)frame_buffer, fb_size, w, h);
	framebuffer_buf_update(fb_size);

}

void* worker_thread(void* arg)
{
    unsigned char *frame = NULL, *tmp = NULL;
    int            frame_size = 0, max_frame_size = 0;
    struct timeval timestamp;


    /* set cleanup handler to cleanup allocated ressources */
    pthread_cleanup_push(worker_cleanup, NULL);

    DBG("-------------OUTPUT FB THREAD START-----------\n");


    fb_size = fb_init();

    while (!pglobal->stop) {

        /* wait for fresh frames */
        pthread_mutex_lock(&pglobal->in[input_number].db);
        pthread_cond_wait(&pglobal->in[input_number].db_update, &pglobal->in[input_number].db);

        /* read buffer */
        frame_size = pglobal->in[input_number].size;

        /* check if framebuffer is large enough, increase it if necessary */
        if (frame_size > max_frame_size) {
            DBG("increasing buffer size to %d\n", frame_size);

            max_frame_size = frame_size + TEN_K;
            if ((tmp = realloc(frame, max_frame_size)) == NULL) {
                free(frame);
                pthread_mutex_unlock(&pglobal->in[input_number].db);
                return NULL;
            }

            frame = tmp;
        }

        /* copy v4l2_buffer timeval to user space */
        timestamp = pglobal->in[input_number].timestamp;
        timestamp = timestamp;

        // ----copy image
        memcpy(frame, pglobal->in[input_number].buf, frame_size);
        DBG("got frame (size: %d kB)\n", frame_size / 1024);

        pthread_mutex_unlock(&pglobal->in[input_number].db);


        // ----update framebuffer
        load_jpeg_to_framebuffer(frame, frame_size);
    }

    free(frame);

    DBG("leaving %s thread, calling cleanup function now\n", OUTPUT_PLUGIN_NAME);
    pthread_cleanup_pop(1);

    return NULL;
}


void worker_cleanup(void* arg)
{
    OPRINT("cleaning up ressources allocated by %s thread\n", OUTPUT_PLUGIN_NAME);
    munmap(fb_addr, fb_size);
    free(fb_buf1);
}


/*************************** (C) COPYRIGHT 2020 ZYS ************END OF FILE****/
